# Copyright (c) 2015-2016, 2019, 2021-2022 by Farsight Security, Inc.
# This work is unpublished, and unauthorized reproduction is prohibited.
# Farsight Security, Inc. owns all rights to this work and intends that
# it be held confidential, preserved as a trade secret, and protected as
# an unpublished copyrighted work.

import heapq
import logging
import os
import time
import threading
import traceback
import urllib.request, urllib.error, urllib.parse
import ctypes
from email.utils import parsedate_tz, mktime_tz
import base64

from .digest import check_digest, digest_extension
from .util import iterfileobj, TempFile

logger = logging.getLogger(__name__)

class TerminatableThread(threading.Thread):
    '''A thread class that supports being terminated from another thread. Updated code from:
    https://github.com/intuited/terminable_thread
    '''
    def terminate(self):
        tid = self.ident
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), ctypes.py_object(SystemExit))
        # Check it
        if res == 0:
            raise ValueError("invalid thread id")
        elif res != 1:
            # "if it returns a number greater than one, you're in trouble,
            # and you should call it again with exc=NULL to revert the effect"
            ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), None)
            raise SystemError("PyThreadState_SetAsyncExc failed")

class DownloadError(Exception): pass

class DownloadManager:
    def __init__(self, max_downloads=4, download_timeout=None, retry_timeout=60):
        self._pending_downloads = set()
        self._active_downloads = dict()

        self._failed_downloads = dict()

        self._max_downloads = max_downloads
        self._download_timeout = download_timeout
        self._retry_timeout = retry_timeout
        self._lock = threading.RLock()
        
        self._main_thread = None
        self._action_required = threading.Condition()
        self._terminate = threading.Event() 

    def start(self):
        logger.debug('Starting DownloadManager {}'.format(self))
        if self._main_thread:
            raise Exception('already running')

        self._terminate.clear()

        self._main_thread = threading.Thread(target=self._run)
        self._main_thread.setDaemon(True)
        self._main_thread.start()

    def stop(self, blocking=False, timeout=None):
        logger.debug('Stopping DownloadManager {}'.format(self))
        self._terminate.set()
        with self._action_required:
            logger.debug('Notifying all run loops')
            self._action_required.notifyAll()
        if blocking or timeout:
            return self.join(timeout=timeout)

    def join(self, timeout=None):
        logger.debug('Joining DownloadManager {}'.format(self))
        return self._main_thread.join(timeout=timeout)

    def _run(self):
        logger.debug('Running DownloadManager {}'.format(self))
        while not self._terminate.is_set():
            with self._lock:
                for f,thread in list(self._active_downloads.items()):
                    if not thread.is_alive():
                        del self._active_downloads[f]
                        logger.debug('Joining {}'.format(thread))
                        thread.join()
                for f,thread in list(self._failed_downloads.items()):
                    if not thread.is_alive():
                        del self._failed_downloads[f]
                        logger.debug('Joining {}'.format(thread))
                        thread.join()

            with self._lock:
                for f in heapq.nlargest(self._max_downloads - len(self._active_downloads), self._pending_downloads):
                    self._pending_downloads.remove(f)

                    thread = TerminatableThread(target=self._download, args=(f,))
                    thread.setDaemon(True)
                    thread.start()

                    self._active_downloads[f] = thread

            with self._action_required:
                logger.debug('Waiting DownloadManager {}'.format(self))
                self._action_required.wait()
                logger.debug('Awoken DownloadManager {}'.format(self))

        logger.debug('Completing DownloadManager run {}'.format(self))
        with self._lock:
            for f,thread in list(self._active_downloads.items()):
                if thread.is_alive():
                    thread.terminate()
                thread.join()
                del self._active_downloads[f]
            for f,thread in list(self._failed_downloads.items()):
                if thread.is_alive():
                    thread.terminate()
                thread.join()
                del self._failed_downloads[f]
        
    def _download(self, f):
        logger.debug('Downloading {}'.format(f))
        try:
            target = f.target()

            logger.info('Downloading {} to {}'.format(f.uri, target))

            req = urllib.request.Request(f.uri)
            if f.apikey:
                req.add_header('X-API-Key', f.apikey)
            fp = urllib.request.urlopen(req, timeout=self._download_timeout)
            out = TempFile(prefix='.{}.'.format(f.name), dir=f.dname)

            algorithm = None
            digest = None
            digest_file = None
            if 'Digest' in fp.headers:
                algorithm,_,digest = fp.headers['Digest'].partition('=')
                digest_file = '{}.{}'.format(target, digest_extension(algorithm))
            elif f.digest_required:
                raise DownloadError('Digest header missing and digest_required=True')

            logger.debug('Copying urlopen of {} to {}'.format(f.uri, out.name))
            for chunk in check_digest(iterfileobj(fp), algorithm, digest):
                out.write(chunk)

            if 'Content-Length' in fp.headers:
                try:
                    expected_len = int(fp.headers['Content-Length'])
                    if out.tell() != expected_len:
                        raise DownloadError('Content length mismatch: {} != {}'.format(out.tell(), expected_len))
                except ValueError:
                    logger.debug('Skipping content length check, invalid header: {}'.format(fp.headers['Content-Length']))
            else:
                logger.debug('Skipping content length check, header missing')

            out.close()
            os.chmod(out.name, 0o644)

            mtime_tz = fp.headers.get('Last-Modified', None)
            if mtime_tz:
                mtime = mktime_tz(parsedate_tz(mtime_tz))
                logger.debug('Setting mtime of {} to {}'.format(out.name, time.ctime(mtime)))
                os.utime(out.name, (mtime, mtime))

            f.validate(out.name)

            if digest_file:
                logger.debug('Writing digest={} to {}'.format(digest, digest_file))
                with TempFile(prefix='.{}.'.format(os.path.basename(digest_file)), dir=f.dname, destination=digest_file) as tf:
                    tf.write('{}  {}\n'.format(base64.b64decode(digest).hex(), os.path.basename(target)).encode('utf-8'))

            try:
                logger.debug('Renaming {} to {}'.format(out.name, target))
                os.rename(out.name, target)
                out.delete = False
            except:
                try:
                    os.unlink(digest_file)
                except OSError:
                    pass
                raise

            logger.info('Download of {} to {} complete'.format(f.uri, target))
        except (KeyboardInterrupt, SystemExit) as e:
            logger.debug('Re-Raising {}'.format(str(e)))
            raise
        except Exception as e:
            logger.error('Download of {} failed: {}'.format(f.uri, str(e)))
            logger.debug(traceback.format_exc())

            expire_thread = TerminatableThread(target=self._expire_failed_download, args=(f,))
            expire_thread.setDaemon(True)
            expire_thread.start()
            with self._lock:
                self._failed_downloads[f] = expire_thread
        finally:
            with self._action_required:
                logger.debug('Notifying run loop')
                self._action_required.notify()

    def _expire_failed_download(self, f, timeout=None):
        if timeout is None:
            timeout = self._retry_timeout
        logger.debug('Waiting {timeout} to retry {uri}'.format(timeout=timeout, uri=f.uri))
        time.sleep(timeout)
        logger.info('Failure timeout for {uri} complete'.format(uri=f.uri))
        with self._action_required:
            logger.debug('Notifying run loop')
            self._action_required.notify()

    def __contains__(self, filename):
        with self._lock:
            return filename in self._pending_downloads or filename in self._active_downloads or filename in self._failed_downloads

    def enqueue(self, f):
        logger.info('Enqueuing {}'.format(os.path.basename(f.name)))

        with self._lock:
            self._pending_downloads.add(f)

        with self._action_required:
            logger.debug('Notifying run loop')
            self._action_required.notify()

