# Copyright (c) 2015-2016, 2019, 2021 by Farsight Security, Inc.
# This work is unpublished, and unauthorized reproduction is prohibited.
# Farsight Security, Inc. owns all rights to this work and intends that
# it be held confidential, preserved as a trade secret, and protected as
# an unpublished copyrighted work.

from io import BytesIO
import base64
import hashlib
import http.client
import os
import tempfile
import unittest
import urllib.request, urllib.parse, urllib.error
import urllib.response

from . import get_uri
from dnstable_manager.download import DownloadManager
from dnstable_manager.fileset import File

class TestDownloadManager(unittest.TestCase):
    @staticmethod
    def noop(self, *args, **kwargs): pass

    def setUp(self):
        self.orig_urlopen = urllib.request.urlopen

    def tearDown(self):
        urllib.request.urlopen = self.orig_urlopen

    def test_download(self):
        tf = tempfile.NamedTemporaryFile(prefix='dns-test-dnstable-manager_download-', suffix='.2015.Y.mtbl', delete=True)
        test_data = 'abc\n123\n'.encode('utf-8')
        digest = base64.b64encode(hashlib.sha256(test_data).digest())
        digest_file = tf.name + '.sha256'
        f = File(os.path.basename(tf.name), dname=os.path.dirname(tf.name))
        f.uri = 'http://example.com/{}'.format(f.name)
        seen_uris = list()
        def my_urlopen(obj, timeout=None):
            uri = get_uri(obj)
            seen_uris.append(uri)
            headers = http.client.HTTPMessage()
            headers.add_header('Content-Length', '{}'.format(len(test_data)))
            headers.add_header('Digest', 'SHA-256={}'.format(digest.decode('utf-8')))
            return urllib.response.addinfourl(BytesIO(test_data), headers, f.uri)
        urllib.request.urlopen = my_urlopen

        m = DownloadManager()
        try:
            m._download(f)
            # assertItemsEqual() was renamed to assertCountEqual() in py3
            self.assertCountEqual(seen_uris, [f.uri])
            with open(tf.name, 'rb') as df:
                self.assertEqual(df.read(), test_data)
            self.assertTrue(os.path.isfile(digest_file), 'Digest file missing: {}'.format(digest_file))
            with open(digest_file) as df:
                test_digest,test_name = df.readline().strip().split()
                self.assertEqual(bytes.fromhex(test_digest), base64.b64decode(digest))
                self.assertEqual(test_name, f.name)
        finally:
            m.stop()
            try:
                os.unlink(digest_file)
            except OSError:
                pass

    def test_download_apikey(self):
        tf = tempfile.NamedTemporaryFile(prefix='dns-test-dnstable-manager_download-', suffix='.2015.Y.mtbl', delete=True)
        test_data = 'abc\n123\n'.encode('utf-8')
        digest = base64.b64encode(hashlib.sha256(test_data).digest())
        digest_file = tf.name + '.sha256'
        apikey = 'TEST API KEY'
        f = File(os.path.basename(tf.name), dname=os.path.dirname(tf.name), apikey=apikey)
        self.assertEqual(f.apikey, apikey)
        f.uri = 'http://example.com/{}'.format(f.name)
        seen_uris = list()
        headers = list()
        def my_urlopen(obj, timeout=None):
            headers.extend(obj.header_items())
            uri = get_uri(obj)
            seen_uris.append(uri)
            response_headers = http.client.HTTPMessage()
            response_headers.add_header('Content-Length', '{}'.format(len(test_data)))
            response_headers.add_header('Digest', 'SHA-256={}'.format(digest.decode('utf-8')))
            return urllib.response.addinfourl(BytesIO(test_data), response_headers, f.uri)
        urllib.request.urlopen = my_urlopen

        m = DownloadManager()
        try:
            m._download(f)
            # assertItemsEqual() was renamed to assertCountEqual() in py3
            self.assertCountEqual(seen_uris, [f.uri])

            for k,v in headers:
                if k.lower() == 'x-api-key':
                    self.assertEqual(v, apikey)
                    break
            else:
                self.fail('X-API-Key header missing')

            with open(tf.name, 'rb') as df:
                self.assertEqual(df.read(), test_data)
            self.assertTrue(os.path.isfile(digest_file), 'Digest file missing: {}'.format(digest_file))
            with open(digest_file) as df:
                test_digest,test_name = df.readline().strip().split()
                self.assertEqual(bytes.fromhex(test_digest), base64.b64decode(digest))
                self.assertEqual(test_name, f.name)
        finally:
            m.stop()
            try:
                os.unlink(digest_file)
            except OSError:
                pass

    def test_download_bad_content_length(self):
        tf = tempfile.NamedTemporaryFile(prefix='dns-test-dnstable-manager_download-', suffix='.2015.Y.mtbl', delete=True)
        test_data = 'abc\n123\n'.encode('utf-8')
        f = File(os.path.basename(tf.name), dname=os.path.dirname(tf.name), digest_required=False)
        f.uri = 'http://example.com/{}'.format(f.name)
        def my_urlopen(obj, timeout=None):
            uri = get_uri(obj)
            self.assertEqual(uri, f.uri)
            headers = http.client.HTTPMessage()
            headers.add_header('Content-Length', '{}'.format(len(test_data)+1))
            return urllib.response.addinfourl(BytesIO(test_data), headers, f.uri)
        urllib.request.urlopen = my_urlopen

        m = DownloadManager()
        try:
            m._download(f)
            self.assertIn(f, m._failed_downloads)
        finally:
            m.stop()

    def test_download_bad_digest(self):
        tf = tempfile.NamedTemporaryFile(prefix='dns-test-dnstable-manager_download-', suffix='.2015.Y.mtbl', delete=True)
        test_data = 'abc\n123\n'.encode('utf-8')
        digest_file = tf.name + '.sha256'
        digest = 'INVALID'
        f = File(os.path.basename(tf.name), dname=os.path.dirname(tf.name))
        f.uri = 'http://example.com/{}'.format(f.name)
        def my_urlopen(obj, timeout=None):
            uri = get_uri(obj)
            self.assertEqual(uri, f.uri)
            headers = http.client.HTTPMessage()
            headers.add_header('Content-Length', '{}'.format(len(test_data)))
            headers.add_header('Digest', 'SHA-256={}'.format(digest))
            return urllib.response.addinfourl(BytesIO(test_data), headers, f.uri)
        urllib.request.urlopen = my_urlopen

        m = DownloadManager()
        try:
            m._download(f)
            self.assertIn(f, m._failed_downloads)
        finally:
            m.stop()
            try:
                os.unlink(digest_file)
            except OSError:
                pass
